
#include "EXTERN.h"
#include "perl.h"
#include "XSUB.h"

#include "ppport.h"
#include "imext.h"
#include "imperl.h"

#include "nv_core.h"
#include "nv_ip.h"
#include "nv_ml.h"
#include "nv_face.h"

#define NV_MAX_FACE 4096

#define NV_C2V(x) ((float)((unsigned int)x))


static void
nv_conv_imager2nv(nv_matrix_t *bgr, nv_matrix_t *gray,
		  i_img *im)
{
  int y, x; 
  i_color c;

  assert(bgr->rows == gray->rows && bgr->rows == im->ysize);
  assert(bgr->cols == gray->cols && bgr->cols == im->xsize);
  
  if (im->channels >= 3) {
    // rgb
    for (y = 0; y < im->ysize; ++y) {
      for (x = 0; x < im->xsize; ++x) {
	i_gpix(im, x, y, &c);
	NV_MAT3D_V(bgr, y, x, NV_CH_B) = NV_C2V(c.rgb.b);
	NV_MAT3D_V(bgr, y, x, NV_CH_G) = NV_C2V(c.rgb.g);
	NV_MAT3D_V(bgr, y, x, NV_CH_R) = NV_C2V(c.rgb.r);
      }
    }
    nv_gray(gray, bgr);
  } else {
    // gray
    for (y = 0; y < im->ysize; ++y) {
      for (x = 0; x < im->xsize; ++x) {
	i_gpix(im, x, y, &c);
	NV_MAT3D_V(gray, y, x, 0) = NV_C2V(c.gray.gray_color);
	NV_MAT3D_V(bgr, y, x, NV_CH_B) = NV_C2V(c.gray.gray_color);
	NV_MAT3D_V(bgr, y, x, NV_CH_G) = NV_C2V(c.gray.gray_color);
	NV_MAT3D_V(bgr, y, x, NV_CH_R) = NV_C2V(c.gray.gray_color);
      }
    }
  }
}


int
_detect(SV *ref,
	i_img *im,
	float step, float scale_factor,
	float min_window_size)
{
  static const nv_mlp_t *detector_mlp = &nv_face_mlp_face_00;
  static const nv_mlp_t *face_mlp[] = {
    &nv_face_mlp_face_01,
    &nv_face_mlp_face_02,
    NULL
  };
  static const nv_mlp_t *dir_mlp = &nv_face_mlp_dir;
  static const nv_mlp_t *parts_mlp = &nv_face_mlp_parts;

  AV *results = (AV *)SvRV(ref);
  
  int i, j, nface;
  nv_rect_t image_size;
  nv_face_position_t face_pos[NV_MAX_FACE];
  
  nv_matrix_t *bgr = nv_matrix3d_alloc(3, im->ysize, im->xsize);
  nv_matrix_t *gray = nv_matrix3d_alloc(1, im->ysize, im->xsize);
  nv_matrix_t *smooth = nv_matrix3d_alloc(1, im->ysize, im->xsize);
  nv_matrix_t *edge = nv_matrix3d_alloc(1, im->ysize, im->xsize);
  nv_matrix_t *gray_integral = nv_matrix3d_alloc(1, im->ysize + 1, im->xsize + 1);
  nv_matrix_t *edge_integral = nv_matrix3d_alloc(1, im->ysize + 1, im->xsize + 1);

  // initialize
  
  nv_matrix_zero(bgr);
  nv_matrix_zero(gray);
  nv_matrix_zero(edge);
  nv_matrix_zero(gray_integral);
  nv_matrix_zero(edge_integral);
  
  image_size.x = image_size.y = 0;
  image_size.width = gray->cols;
  image_size.height = gray->rows;
  
  // convert format
  nv_conv_imager2nv(bgr, gray, im);
  // edge
  nv_gaussian5x5(smooth, 0, gray, 0);
  nv_laplacian1(edge, smooth, 4.0f);
  // integral
  nv_integral(gray_integral, gray, 0);
  nv_integral(edge_integral, edge, 0);

  // detect face
  nface = nv_face_detect(face_pos, NV_MAX_FACE,
			 gray_integral, edge_integral, &image_size,
			 dir_mlp,
			 detector_mlp, face_mlp, 2,
			 parts_mlp,
			 step, scale_factor, min_window_size
			 );
  // analyze face 
  for (i = 0; i < nface; ++i) {
    nv_face_feature_t face_feature = {0};
    nv_face_analyze(&face_feature, &face_pos[i], bgr);

    /*
     * likelihood => ,
     * face => {x=>,y=>,width=>,height=>, skin_color =>, hair_color=>, },
     * eyes => {left} => { x=>,y=>,width=>,height=> , color => (,,,,)},
     * eyes => {right} => { x=>,y=>,width=>,height=> , color => (,,,,)}, 
     * nose => {x=>, y=>, width=>1, height=>1},
     * mouse => { x=>, y=>, width=>, height=>},
     * chin => { x=>, y =>, width=>1, height=>1}
     */
    
    // likelihood
    av_push(results, newSVnv(face_pos[i].likelihood));
    
    // face.x,y,w,h,skin_color(b,g,r),hair_color(b,g,r)
    av_push(results, newSViv(face_pos[i].face.x));
    av_push(results, newSViv(face_pos[i].face.y));
    av_push(results, newSViv(face_pos[i].face.width));
    av_push(results, newSViv(face_pos[i].face.height));
    av_push(results, newSViv(face_feature.skin_bgr.v[0]));
    av_push(results, newSViv(face_feature.skin_bgr.v[1]));
    av_push(results, newSViv(face_feature.skin_bgr.v[2]));
    av_push(results, newSViv(face_feature.hair_bgr.v[0]));
    av_push(results, newSViv(face_feature.hair_bgr.v[1]));
    av_push(results, newSViv(face_feature.hair_bgr.v[2]));

    // left_eye x,y,w,h,color[4](b,g,r)
    av_push(results, newSViv(face_pos[i].left_eye.x));
    av_push(results, newSViv(face_pos[i].left_eye.y));
    av_push(results, newSViv(face_pos[i].left_eye.width));
    av_push(results, newSViv(face_pos[i].left_eye.height));
    for (j = 0; j < 4; ++j) {
      av_push(results, newSViv(face_feature.left_eye_bgr[j].v[0]));
      av_push(results, newSViv(face_feature.left_eye_bgr[j].v[1]));
      av_push(results, newSViv(face_feature.left_eye_bgr[j].v[2]));
    }

    // right_eye x,y,w,h,color[4](b,g,r)
    av_push(results, newSViv(face_pos[i].right_eye.x));
    av_push(results, newSViv(face_pos[i].right_eye.y));
    av_push(results, newSViv(face_pos[i].right_eye.width));
    av_push(results, newSViv(face_pos[i].right_eye.height));
    for (j = 0; j < 4; ++j) {
      av_push(results, newSViv(face_feature.right_eye_bgr[j].v[0]));
      av_push(results, newSViv(face_feature.right_eye_bgr[j].v[1]));
      av_push(results, newSViv(face_feature.right_eye_bgr[j].v[2]));
    }

    // nose x,y
    av_push(results, newSViv(face_pos[i].nose.x));
    av_push(results, newSViv(face_pos[i].nose.y));
    
    // mouth x,y,w,h
    av_push(results, newSViv(face_pos[i].mouth.x));
    av_push(results, newSViv(face_pos[i].mouth.y));
    av_push(results, newSViv(face_pos[i].mouth.width));
    av_push(results, newSViv(face_pos[i].mouth.height));

    // chin x, y
    av_push(results, newSViv(face_pos[i].chin.x));
    av_push(results, newSViv(face_pos[i].chin.y));
  }
  nv_matrix_free(&bgr);
  nv_matrix_free(&gray);
  nv_matrix_free(&smooth);
  nv_matrix_free(&edge);
  nv_matrix_free(&gray_integral);
  nv_matrix_free(&edge_integral);

  return nface;
}


DEFINE_IMAGER_CALLBACKS;

MODULE = Imager::AnimeFace   PACKAGE = Imager::AnimeFace

PROTOTYPES: ENABLE

int _detect(ref, im, step, scale_factor, min_window_size)
  SV *ref
  Imager::ImgRaw im
  float step
  float scale_factor
  float min_window_size
 
BOOT:
  PERL_INITIALIZE_IMAGER_CALLBACKS;
